package org.kafkahq.repositories;

import com.google.common.collect.ImmutableMap;
import org.apache.kafka.clients.admin.ConfigEntry;
import org.apache.kafka.common.config.ConfigResource;
import org.kafkahq.modules.KafkaModule;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.stream.Collectors;

@Singleton
public class ConfigRepository extends AbstractRepository {
    @Inject
    private KafkaModule kafkaModule;

    public List<org.kafkahq.models.Config> findByBroker(Integer name) throws ExecutionException, InterruptedException {
        return this.findByBrokers(Collections.singletonList(name)).get(String.valueOf(name));
    }

    public Map<String, List<org.kafkahq.models.Config>> findByBrokers(List<Integer> names) throws ExecutionException, InterruptedException {
        return this.find(
            ConfigResource.Type.BROKER,
            names
                .stream()
                .map(String::valueOf)
                .collect(Collectors.toList())
        );
    }

    public List<org.kafkahq.models.Config> findByTopic(String name) throws ExecutionException, InterruptedException {
        return this.findByTopics(Collections.singletonList(name)).get(name);
    }

    public Map<String, List<org.kafkahq.models.Config>> findByTopics(List<String> names) throws ExecutionException, InterruptedException {
        return this.find(ConfigResource.Type.TOPIC, names);
    }

    private Map<String, List<org.kafkahq.models.Config>> find(ConfigResource.Type type, List<String> names) throws ExecutionException, InterruptedException {
        Map<String, List<org.kafkahq.models.Config>> map = new HashMap<>();

        kafkaWrapper.describeConfigs(type, names).forEach((key, value) -> {
            List<org.kafkahq.models.Config> collect = value.entries()
                .stream()
                .map(org.kafkahq.models.Config::new)
                .sorted(Comparator.comparing(org.kafkahq.models.Config::getName))
                .collect(Collectors.toList());

            map.put(key.name(), collect);
        });

        return map;
    }

    public void updateBroker(String clusterId, Integer name, List<org.kafkahq.models.Config> configs) throws ExecutionException, InterruptedException {
        this.update(clusterId, ConfigResource.Type.BROKER, String.valueOf(name), configs);
    }

    public void updateTopic(String clusterId, String name, List<org.kafkahq.models.Config> configs) throws ExecutionException, InterruptedException {
        this.update(clusterId, ConfigResource.Type.TOPIC, name, configs);
    }

    private void update(String clusterId, ConfigResource.Type type, String name, List<org.kafkahq.models.Config> configs) throws ExecutionException, InterruptedException {
        List<ConfigEntry> entries = configs
            .stream()
            .map(config -> new ConfigEntry(config.getName(), config.getValue()))
            .collect(Collectors.toList());

        kafkaModule.getAdminClient(clusterId)
            .alterConfigs(ImmutableMap.of(
                new ConfigResource(type, name),
                new org.apache.kafka.clients.admin.Config(entries)
            ))
            .all()
            .get();
    }

    public static List<org.kafkahq.models.Config> updatedConfigs(Map<String, String> request, List<org.kafkahq.models.Config> configs) {
        return configs
            .stream()
            .filter(config -> !config.isReadOnly())
            .filter(config -> request.containsKey("configs[" + config.getName() + "]"))
            .filter(config -> {
                String current = config.getValue() == null ? "" : config.getValue();

                return !(current).equals(request.get("configs[" + config.getName() + "]"));
            })
            .map(config -> config.withValue(request.get("configs[" + config.getName() + "]")))
            .collect(Collectors.toList());
    }
}
