package org.kafkahq.clusters;

import io.confluent.kafka.schemaregistry.RestApp;
import io.confluent.kafka.schemaregistry.avro.AvroCompatibilityLevel;
import io.confluent.kafka.schemaregistry.rest.SchemaRegistryConfig;
import kafka.server.KafkaConfig$;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.extension.AfterTestExecutionCallback;
import org.junit.jupiter.api.extension.BeforeTestExecutionCallback;
import org.junit.jupiter.api.extension.ExtensionContext;

import java.io.IOException;
import java.net.ServerSocket;
import java.util.Properties;

@Slf4j
public class EmbeddedSingleNodeKafkaCluster implements BeforeTestExecutionCallback, AfterTestExecutionCallback {
    private static final int DEFAULT_BROKER_PORT = 0; // 0 results in a random port being selected
    private static final String KAFKA_SCHEMAS_TOPIC = "__schemas";
    private static final String AVRO_COMPATIBILITY_TYPE = AvroCompatibilityLevel.BACKWARD.name;

    private static final String KAFKASTORE_OPERATION_TIMEOUT_MS = "10000";
    private static final String KAFKASTORE_DEBUG = "true";
    private static final String KAFKASTORE_INIT_TIMEOUT = "90000";

    private ZooKeeperEmbedded zookeeper;
    private KafkaEmbedded broker;
    private RestApp schemaRegistry;
    private ConnectEmbedded connect;
    private final Properties brokerConfig;
    private boolean running;

    public EmbeddedSingleNodeKafkaCluster(final Properties brokerConfig) {
        this.brokerConfig = new Properties();
        this.brokerConfig.putAll(brokerConfig);
    }

    public void start() throws Exception {
        // zookeeper
        log.debug("Initiating embedded Kafka cluster startup");
        log.debug("Starting a ZooKeeper instance...");
        zookeeper = new ZooKeeperEmbedded();
        log.debug("ZooKeeper instance is running at {}", zookeeper.connectString());

        // kafka
        final Properties effectiveBrokerConfig = effectiveBrokerConfigFrom(brokerConfig, zookeeper);
        log.debug("Starting a Kafka instance on port {} ...", effectiveBrokerConfig.getProperty(KafkaConfig$.MODULE$.PortProp()));
        broker = new KafkaEmbedded(effectiveBrokerConfig);
        log.debug("Kafka instance is running at {}, connected to ZooKeeper at {}", broker.brokerList(), broker.zookeeperConnect());

        // schema registry
        final Properties schemaRegistryProps = new Properties();
        schemaRegistryProps.put(SchemaRegistryConfig.KAFKASTORE_TIMEOUT_CONFIG, KAFKASTORE_OPERATION_TIMEOUT_MS);
        schemaRegistryProps.put(SchemaRegistryConfig.DEBUG_CONFIG, KAFKASTORE_DEBUG);
        schemaRegistryProps.put(SchemaRegistryConfig.KAFKASTORE_INIT_TIMEOUT_CONFIG, KAFKASTORE_INIT_TIMEOUT);

        schemaRegistry = new RestApp(0, zookeeperConnect(), KAFKA_SCHEMAS_TOPIC, AVRO_COMPATIBILITY_TYPE, schemaRegistryProps);
        schemaRegistry.start();
        log.debug("Schema registry is running at {}", schemaRegistryUrl());

        // connect
        ServerSocket s = new ServerSocket(0);

        Properties properties = new Properties();
        properties.put("bootstrap.servers", bootstrapServers());
        properties.put("key.converter", "io.confluent.connect.avro.AvroConverter");
        properties.put("key.converter.schema.registry.url", schemaRegistryUrl());
        properties.put("value.converter", "io.confluent.connect.avro.AvroConverter");
        properties.put("value.converter.schema.registry.url", schemaRegistryUrl());
        properties.put("rest.port", "0");
        properties.put("group.id", "connect-integration-test-");
        properties.put("offset.storage.topic", "__connect-offsets");
        properties.put("offset.storage.replication.factor", 1);
        properties.put("config.storage.topic", "__connect-config");
        properties.put("config.storage.replication.factor", 1);
        properties.put("status.storage.topic", "__connect-status");
        properties.put("status.storage.replication.factor", 1);

        connect = new ConnectEmbedded(properties);
        log.debug("Kafka Connect is running at {}", connectUrl());

        running = false;
    }

    private Properties effectiveBrokerConfigFrom(final Properties brokerConfig, final ZooKeeperEmbedded zookeeper) {
        final Properties effectiveConfig = new Properties();
        effectiveConfig.putAll(brokerConfig);
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkConnectProp(), zookeeper.connectString());
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkSessionTimeoutMsProp(), 30 * 1000);
        effectiveConfig.put(KafkaConfig$.MODULE$.PortProp(), DEFAULT_BROKER_PORT);
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkConnectionTimeoutMsProp(), 60 * 1000);
        effectiveConfig.put(KafkaConfig$.MODULE$.DeleteTopicEnableProp(), true);
        effectiveConfig.put(KafkaConfig$.MODULE$.LogCleanerDedupeBufferSizeProp(), 2 * 1024 * 1024L);
        effectiveConfig.put(KafkaConfig$.MODULE$.GroupMinSessionTimeoutMsProp(), 0);
        effectiveConfig.put(KafkaConfig$.MODULE$.OffsetsTopicReplicationFactorProp(), (short) 1);
        effectiveConfig.put(KafkaConfig$.MODULE$.OffsetsTopicPartitionsProp(), 1);
        effectiveConfig.put(KafkaConfig$.MODULE$.AutoCreateTopicsEnableProp(), true);
        return effectiveConfig;
    }

    @Override
    public void beforeTestExecution(ExtensionContext context) throws Exception {
        start();
    }

    @Override
    public void afterTestExecution(ExtensionContext context) throws Exception {
        stop();
    }

    public void stop() {
        log.info("Stopping EmbeddedSingleNodeKafkaCluster");
        try {
            try {
                if (connect != null) {
                    connect.stop();
                }
            } catch (final Exception e) {
                throw new RuntimeException(e);
            }

            try {
                if (schemaRegistry != null) {
                    schemaRegistry.stop();
                }
            } catch (final Exception e) {
                throw new RuntimeException(e);
            }

            if (broker != null) {
                broker.stop();
            }

            try {
                if (zookeeper != null) {
                    zookeeper.stop();
                }
            } catch (final IOException e) {
                throw new RuntimeException(e);
            }
        } finally {
            running = false;
        }
        log.info("EmbeddedSingleNodeKafkaCluster Stopped");
    }

    public String bootstrapServers() {
        return broker.brokerList();
    }

    public String zookeeperConnect() {
        return zookeeper.connectString();
    }

    public String schemaRegistryUrl() {
        return schemaRegistry.restConnect;
    }

    public String connectUrl() {
        return connect.connectUrl();
    }
}